@@ -7,14 +7,20 @@ import json
 import threading
 import time
 import getopt
+import shlex
+
+from typing import Tuple, Union
 
 from tempfile import gettempdir
 from crontab import CronTab
+from dotenv import dotenv_values
+from io import StringIO
 
 from containerd.services.containers.v1 import containers_pb2_grpc, containers_pb2
 from containerd.services.events.v1 import unwrap, events_pb2, events_pb2_grpc
 from containerd.services.tasks.v1 import tasks_pb2, tasks_pb2_grpc
 
+from apscheduler.schedulers.base import BaseScheduler
 from apscheduler.schedulers.background import BackgroundScheduler
 from apscheduler.executors.pool import ThreadPoolExecutor
 from apscheduler.triggers.cron import CronTrigger
@@ -36,6 +42,7 @@ def hashsum(str):
     md5.update(str.encode('utf-8'))
     return md5.hexdigest()
 
+
 class ReadTask(threading.Thread):
     abort_event: threading.Event
     output: bytearray
@@ -77,15 +84,20 @@ def mkfifo(exec_id: str):
     return fifo
 
 
-def run_command(channel, container_id, args, timeout=5):
+def get_container_spec(channel: grpc.Channel, container_id: str):
+    containers_stub = containers_pb2_grpc.ContainersStub(channel)
+    container = containers_stub.Get(containers_pb2.GetContainerRequest(
+        id=container_id), metadata=METADATA).container
+    return json.loads(container.spec.value)
+
+
+def run_command(channel: grpc.Channel, container_id: str, args, timeout=5) -> Tuple[int, str]:
     exec_id = "exec-" + hashsum(container_id + ': ' + str(args))
     stdout = mkfifo(exec_id)
     read_task = ReadTask(stdout)
 
-    containers_stub = containers_pb2_grpc.ContainersStub(channel)
-    container = containers_stub.Get(containers_pb2.GetContainerRequest(
-        id=container_id), metadata=METADATA).container
-    container_process = json.loads(container.spec.value)['process']
+    container_spec = get_container_spec(channel, container_id)
+    container_process = container_spec['process']
 
     process = {
         'args': args,
@@ -130,27 +142,80 @@ def run_command(channel, container_id, args, timeout=5):
     return exit_status, read_task.result()
 
 
-def run_schedule(channel, container_id, args):
-    exit_code = run_command(channel, container_id, args)
+def run_schedule(channel: grpc.Channel, container_id: str, args):
+    exit_code, _ = run_command(channel, container_id, args)
     logging.getLogger('cron').info(
-        "{code} <- schedule {schedule}".format(code=exit_code, schedule=' '.join(args)))
+        "{code} <- schedule {schedule}".format(code=exit_code, schedule=' '.join(shlex.quote(arg) for arg in args)))
+    return exit_code
 
 
-def load_container_schedules(scheduler, container_id, channel):
-    logging.getLogger('cron').info(
-        'load schedules from {container_id}'.format(container_id=container_id))
-
-    args = ["/bin/sh", "-c",
-            '[ -d /etc/cron.d ] && find /etc/cron.d ! -name \".*\" -type f -exec cat \{\} \;']
-    exit_code, output = run_command(channel, container_id, args)
+def get_os_id(channel: grpc.Channel, container_id: str):
+    exit_code, output = run_command(
+        channel, container_id, ["cat", "/etc/os-release"])
     if exit_code != 0:
         return
+    release = dotenv_values(stream=StringIO(output.decode('utf8')))
+    return release['ID']
 
-    tab = output.decode('utf8').replace('\t', ' ')
+
+def get_alpine_crontab(channel: grpc.Channel, container_id: str) -> Tuple[str, Union[str, bool]]:
+    exit_code, output = run_command(
+        channel, container_id, ['whoami'])
+    if exit_code != 0:
+        return '', None
+    user = output.decode('utf8').replace('\n', '')
+    exit_code, output = run_command(
+        channel, container_id, ['cat', '/etc/crontabs/{user}'.format(user=user)])
+    if exit_code != 0:
+        return '', None
+    return output.decode('utf8').replace('\t', ' '), user
+
+
+def get_debian_crontab(channel: grpc.Channel, container_id: str) -> Tuple[str, Union[str, bool]]:
+    exit_code, output = run_command(channel, container_id, ["/bin/sh", "-c",
+                                                            '[ -d /etc/cron.d ] && find /etc/cron.d ! -name \".*\" -type f -exec cat \{\} \;'])
+    if exit_code != 0:
+        return '', None
+    return output.decode('utf8').replace('\t', ' '), False
+
+
+def get_container_crontab(channel: grpc.Channel, container_id: str) -> Tuple[str, Union[str, bool]]:
+    os_id = get_os_id(channel, container_id)
+    if os_id == 'alpine':
+        return get_alpine_crontab(channel, container_id)
+    elif os_id == 'debian' or os_id == 'ubuntu':
+        return get_debian_crontab(channel, container_id)
+    return '', None
+
+
+def parse_args(command: str):
+    return shlex.split(command)
+
+
+def get_container_name(spec):
+    if 'annotations' in spec:
+        annotations = spec['annotations']
+        if 'io.kubernetes.cri.container-name' in annotations:
+            return annotations['io.kubernetes.cri.container-name']
+    if 'hostname' in spec:
+        return spec['hostname']
+    if 'id' in spec:
+        return spec['id']
+
+
+def load_container_schedules(scheduler: BaseScheduler, container_id, channel):
+    tab, user = get_container_crontab(channel, container_id)
     if tab == '':
         return
 
-    cron_jobs = CronTab(tab=tab, user=False)
+    logger = logging.getLogger('cron')
+    container_spec = get_container_spec(channel, container_id)
+    container_name = get_container_name(
+        container_spec) if not None else container_id
+    logger.info(
+        'load schedules from [{container_name}]...'.format(container_name=container_name))
+
+    cron_jobs = CronTab(tab=tab, user=user)
     for job in cron_jobs:
         if not job.is_enabled():
             continue
@@ -159,31 +224,40 @@ def load_container_schedules(scheduler, container_id, channel):
             slices = SPECIALS[slices.lstrip('@')]
         scheduler.add_job(run_schedule,
                           CronTrigger.from_crontab(slices),
-                          args=[channel, container_id, job.command],
+                          args=[channel, container_id,
+                                parse_args(job.command)],
                           name=job.command)
+        logger.debug(
+            'found [{container_name}]: {job}.'.format(job=job.command, container_name=container_name))
+
+    logger.info(
+        'got {job_count} schedules now.'.format(job_count=len(scheduler.get_jobs())))
 
 
-def unload_container_schedules(scheduler, container_id):
-    logging.getLogger('cron').info(
-        'unload schedules from {container_id}'.format(container_id=container_id))
-
-    for job in scheduler.get_jobs():
+def unload_container_schedules(scheduler: BaseScheduler, container_id):
+    jobs = scheduler.get_jobs()
+    job_count = len(jobs)
+    for job in jobs:
         # 若存储器中的任务所属容器当前不存在，则在存储请中删除此任务
-        if job.args[0] == container_id:
+        if job.args[1] == container_id:
             scheduler.remove_job(job_id=job.id)
+            job_count -= 1
+    logging.getLogger('cron').info(
+        'some schedules removed, {job_count} left.'.format(job_count=job_count))
 
 
 def main():
     try:
-        opts, args = getopt.getopt(sys.argv[1:], 's:n:', ['cri-socket=', 'namespace='])
+        opts, _ = getopt.getopt(sys.argv[1:], 's:n:', [
+            'cri-socket=', 'namespace='])
     except getopt.GetoptError as e:
         print('Usage: --cri-socket|-s <SOCKET> --namespace|-n <NAMESPACE>')
         exit(1)
 
     cri_socket = 'unix:///run/containerd/containerd.sock'
-    namespace = 'k8s.io' # moby for docker
+    namespace = 'k8s.io'  # moby for docker
 
-    for k,v in opts:
+    for k, v in opts:
         if k == '--cri-socket' or k == '-s':
             cri_socket = 'unix://' + v if v.startswith('/') else v
         elif k == '--namespace' or k == '-n':
@@ -195,15 +269,12 @@ def main():
     TIMEZONE = os.getenv('TIMEZONE', 'Asia/Shanghai')
 
     logging.basicConfig(stream=sys.stdout)
-    logging.getLogger('apscheduler').setLevel(logging.INFO)
+    logging.getLogger('apscheduler').setLevel(logging.ERROR)
     logging.getLogger('cron').setLevel(logging.INFO)
 
     scheduler = BackgroundScheduler(
         executors={'default': ThreadPoolExecutor(40)}, timezone=TIMEZONE)
-    try:
-        scheduler.start()
-    except:
-        pass
+    scheduler.start()
 
     with grpc.insecure_channel(cri_socket) as channel:
         containers_stub = containers_pb2_grpc.ContainersStub(channel)
